!
! IDRS Induced Dimension Reduction method
!
!   x = IDRS( A, b, M, s,                             (mandatory)
!           [ tolerance, maximum_iterations, variant, (optional input scalars)
!             flag, relres, iterations,               (optional output scalars)
!             x0, U0, omega,                          (optional input vectors)
!             resvec, H] )                            (optional output vectors)
!
!   solves the system of linear equations A*x=b for x with the method IDRS
!   The dimension of the matrix A is NxN 
!   The dimension of the right-hand-side b is NxNRHS 
!   (NRHS: number of rhs vectors)
!
!   IDRS uses a user defined TYPE MATRIX, which is a structure that contains
!   all the parameters that define the system matrix A
!   The operator * must be overloaded to perform the operation A*v
!   To this end overloading must be defined for (real and complex) user 
!   routines that perform the matrix-vector multiplication. These
!   routines should be included in the module user_module that is used in the
!   module idrs_module.f90.
!
!   IDRS uses a user defined TYPE PRECONDITIONER, which is a structure that 
!   contains all the parameters that define the preconditioning matrix M
!   Overloading of the operator / is used to perform the preconditioning 
!   operation v/M1. To this end overloading must be defined for (real and 
!   complex) user routines that perform the preconditiong operations. These 
!   routines should be included in the module user_module that is used in the
!   module idrs_module.f90.
!
!   The precision of the complex and real variables are defined by the 
!   parameters rp (real precision)and cp (complex precision). These parameters 
!   should be set in the module user_module.
!
!   Input and output variables of IDRS:
!
!   Function output:
!        x: REAL (kind=rp) or COMPLEX (kind=cp) vector of dimension NxNRHS
!         
!   Function parameters:
!     Input, required:
!        A: TYPE MATRIX (user defined), A defines the system matrix 
!        b: REAL (kind=rp) or COMPLEX (kind=cp) matrix of dimension NxNRHS,
!           b is the matrix of right-hand-side vectors
!        M: TYPE PRECONDITIONER, M defines the preconditioner
!        s: INTEGER (must be > 0), s defines the dimension of the shadow space 
!           in IDRS, and by this the depth of the recursions 
!
!     Input, optional parameters:
!        tolerance: REAL (kind=rp). IDRS is terminated if |b-Ax|/|b| < tolerance
!           Default: tolerance = 1e-6
!        maximum_iterations: INTEGER. IDRS stops of the number of iterations 
!           exceeds maximum_iterations. 
!           Default: maximum_iterations = min(2*N,1000)
!        variant: CHARACTER*8. Selects the specific IDR-variant. 
!           Possible variants are:
!           variant = 'biortho ': IDR(s) with random shadow vectors,
!                                 'maintaining convergence' for computing omega
!           variant = 'minsync ': Piecewise constant shadow vectors, 
!                                 'minimal residual' for computing omega
!           variant = 'bicgstab': s must be set to 1. 
!                                 Initial residual is the shadow vector,
!                                 'minimal residual' for omega. 
!                                 This gives an algorithm that is
!                                 mathematically the same as BiCGSTAB
!           Default: variant = 'biortho '
!
!     Output, optional parameters:
!        flag: INTEGER. Indicates convergence condition
!           flag = 0: required tolerance satisfied
!           flag = 1: no convergence to the required tolerance within maximum
!                     number of iterations
!           flag = 2: final residual norm above tolerance
!           flag = 3: one of the iteration parameters became zero, 
!                     causing break down
!        relres: REAL (kind=rp), |b-Ax|/|b|. The norms are trace norms.
!        iterations: INTEGER. The number of iterations that has been performed 
!                             to compute the solution
!
!     Input, optional arrays:
!        x0: REAL (kind=rp) or COMPLEX (kind=cp) matrix of dimension NxNRHS,
!           initial guess for the solution. Default: x0 = 0.
!        U0: REAL (kind=rp) or COMPLEX (kind=cp) array of dimension NxNRHSxs,
!           Initial search space. Default: no initial search space.
!        omega: REAL (kind=rp) or COMPLEX (kind=cp) array of dimension n_omega,
!           where n_omega is the number of user defined omega-parameters.
!           The parameters are used cyclicly. User defined omega's make it 
!           possible to define and examine new IDR(s) variants.
!
!     Output, optional arrays:
!        resvec: REAL (kind=rp), array of dimension maximum_iterations+1
!           resvec contains for every iteration the norm of the residual
!        H: REAL (kind=rp) or COMPLEX (kind=cp) matrix of dimension 
!           (nritz+1) x nritz, in which nritz is the number of ritz values
!           that can be computed. Note that in s+1 iterations s additional 
!           ritz values can be computed.  H is an extended upper Hessenberg 
!           matrix of which the upper bandwidth is s. It can be used for 
!           analysis purposes, the eigenvalues of the leading submatrices of 
!           H are ritzvalues
!
!   Acknowledgement: Duncan van der Heul, Reinaldo Astudillo, Jan de Gier and 
!                    Marielba Rojas are gratefully acknowledged for their 
!                    advise on many aspects of this software. 
!
!   The software is distributed without any warranty.
!
!   Martin van Gijzen
!   Copyright (c) July 2015
!

module idrs_module

   use user_module

   implicit none

   private
   public :: IDRS, TRACE_DOT, FROB_NORM, P_DOT

   interface TRACE_DOT
      module procedure CTRACE_DOT, RTRACE_DOT, RCTRACE_DOT
   end interface

   interface P_DOT
      module procedure CP_DOT, RP_DOT
   end interface

   interface FROB_NORM
      module procedure CFROB_NORM, RFROB_NORM
   end interface

   interface IDRS
      module procedure CIDRS, RIDRS
   end interface

contains

   function CTRACE_DOT(v,w)

! Trace inner product of complex matrices

      complex(kind=cp), intent(in)      :: v(:,:), w(:,:)
      complex(kind=cp)                  :: CTRACE_DOT

      CTRACE_DOT = sum( conjg(v)*w )

   end function CTRACE_DOT

   function RTRACE_DOT(v,w)

! Trace inner product of real matrices

      real(kind=rp), intent(in)      :: v(:,:), w(:,:)
      real(kind=rp)                  :: RTRACE_DOT

      RTRACE_DOT = sum( v*w ) 

   end function RTRACE_DOT

   function RCTRACE_DOT(v,w)

! Trace inner product of real matrices

      real(kind=rp), intent(in)      :: v(:,:)
      complex(kind=cp)               :: w(:,:)
      complex(kind=cp)               :: RCTRACE_DOT

      RCTRACE_DOT = sum( v*w )

   end function RCTRACE_DOT

   function CP_DOT(P,R0,w,s)

! P inner product of complex matrices

      integer                                   :: s
      real(kind=rp),    allocatable, intent(in) :: P(:,:,:) 
      complex(kind=cp), allocatable, intent(in) :: R0(:,:)
      complex(kind=cp), intent(in)              :: w(:,:)
      complex(kind=cp)                          :: v(s), CP_DOT(s)
      integer                                   :: i, j, N, low(s), up(s), step, nrhs

      if ( allocated(P) ) then
! Biortho: P has orthogonal random numbers
         do i = 1,s
            v(i) = sum( P(:,:,i)*w )
         end do
      else if ( allocated(R0) ) then
! BiCGSTAB: shadow vector equal to initial residual
          v(1) = sum( conjg(R0)*w )
      else
! Minsync: P is piecewise constant 
         N    = size(w,1)
         nrhs = size(w,2)
         step = N/s
         low(1) = 1
         do i = 1,s-1
            low(i+1) = i*step+1
            up(i) = i*step
         end do
         up(s) = N

         do i = 1,s
            v(i)  = 0.
            do j = 1, nrhs
               v(i) = v(i) + sum( w(low(i):up(i),j) )
            end do
         end do
      end if
      
      CP_DOT = v

   end function CP_DOT

   function RP_DOT(P,R0,w,s)

! P inner product of real matrices

      integer                              :: s
      real(kind=rp),allocatable,intent(in) :: P(:,:,:) 
      real(kind=rp),allocatable,intent(in) :: R0(:,:) 
      real(kind=rp), intent(in)            :: w(:,:)
      real(kind=rp)                        :: v(s), RP_DOT(s)
      integer                              :: i, j, N, nrhs, low(s), up(s), step

      if ( allocated(P) ) then
! Biortho: P has orthogonal random numbers
         do i = 1,s
            v(i) = sum( P(:,:,i)*w )
         end do
      else if (allocated(R0) ) then
! Bi-CGSTAB: shadow vector equal to initial residual
         v(1) = sum( R0*w )
      else
! Minsync: P has piecewise constant columns
         N = size(w,1)
         nrhs = size(w,2)
         step = N/s
         low(1) = 1
         up(s) = N
         do i = 1,s-1
            low(i+1) = i*step+1
            up(i) = i*step
         end do

         do i = 1,s
            v(i) = 0.
            do j = 1, nrhs
               v(i) = sum( w(low(i):up(i),j) )
            end do
         end do
      end if
         
      RP_DOT = v

   end function RP_DOT 

   function CFROB_NORM(v)

! Frobenius norm of complex matrix

      complex(kind=cp), intent(in)      :: v(:,:)
      real(kind=rp)                     :: CFROB_NORM

      CFROB_NORM = sqrt( sum( conjg(v)*v ) )

   end function CFROB_NORM

   function RFROB_NORM(v)

! Frobenius norm of complex matrix

      real(kind=rp), intent(in)      :: v(:,:)
      real(kind=rp)                  :: RFROB_NORM

      RFROB_NORM = sqrt( sum( v**2 ) )

   end function RFROB_NORM

   function CIDRS( A, b, M1, s, &                            ! required
                   tolerance, maximum_iterations, variant, & ! optional input
                   flag, relres, iterations, &               ! optional output
                   x0, U0, omega, resvec, H )                ! optional arrays

   IMPLICIT NONE

! Required input parameters:
   type(matrix), intent(in)               :: A               ! system matrix
   complex(kind=cp), intent(in)           :: b(:,:)          ! system rhs
   type(preconditioner), intent(in)       :: M1              ! preconditioner
   integer, intent(in)                    :: s               ! s parameter

! Solution:
   complex(kind=cp)                       :: CIDRS(size(b,1),size(b,2))

! Optional input parameters:
   real(kind=rp), optional, intent(in)    :: tolerance 
   integer, optional, intent(in)          :: maximum_iterations 
   character(len=8), optional, intent(in) :: variant     

! Optional output parameters:
   integer, optional, intent(out)         :: flag        
   real(kind=rp), optional, intent(out)   :: relres     
   integer, optional, intent(out)         :: iterations

! Optional input arrays:
   complex(kind=cp), optional, intent(in) :: x0(:,:)  
   complex(kind=cp), optional, intent(in) :: U0(:,:,:) 
   complex(kind=cp), optional, intent(in) :: omega(:) 

! Optional output arrays:
   real(kind=rp), optional, intent(out)   :: resvec(:) 
   complex(kind=cp), optional, intent(out):: H(:,:)   
    
! Local arrays:
   real(kind=rp),    allocatable          :: P(:,:,:) 
   complex(kind=cp), allocatable          :: R0(:,:) 
   complex(kind=cp)                       :: x(size(b,1),size(b,2))
   complex(kind=cp)                       :: G(size(b,1),size(b,2),s)
   complex(kind=cp)                       :: U(size(b,1),size(b,2),s)
   complex(kind=cp)                       :: r(size(b,1),size(b,2)) 
   complex(kind=cp)                       :: v(size(b,1),size(b,2))   
   complex(kind=cp)                       :: t(size(b,1),size(b,2))  
   complex(kind=cp)                       :: M(s,s), f(s), mu(s)
   complex(kind=cp)                       :: alpha(s), beta(s), gamma(s)
   
   complex(kind=cp)                       :: om, tr    
   real(kind=rp)                          :: nr, nt, rho, kappa

   include 'idrs_body.f90'
   cidrs = x
   
   end function CIDRS

   function RIDRS( A, b, M1, s, &                            ! required
                   tolerance, maximum_iterations, variant, & ! optional input
                   flag, relres, iterations, &               ! optional output
                   x0, U0, omega, H, resvec )                ! optional arrays

   IMPLICIT NONE

! Required input and output parameters:
   type(matrix), intent(in)               :: A               ! system matrix
   real(kind=rp), intent(in)              :: b(:,:)          ! system rhs
   type(preconditioner), intent(in)       :: M1              ! preconditioner
   integer, intent(in)                    :: s               ! s parameter

! Solution
   real(kind=rp)                          :: RIDRS(size(b,1),size(b,2))

! Optional input parameters:
   real(kind=rp), optional, intent(in)    :: tolerance
   integer, optional, intent(in)          :: maximum_iterations
   character(len=8), optional, intent(in) :: variant

! Optional output parameters:
   integer, optional, intent(out)         :: flag
   real(kind=rp), optional, intent(out)   :: relres
   integer, optional, intent(out)         :: iterations 

! Optional input arrays:
   real(kind=rp), optional, intent(in)    :: x0(:,:)
   real(kind=rp), optional, intent(in)    :: U0(:,:,:)
   real(kind=rp), optional, intent(in)    :: omega(:)

! Optional output arrays
   real(kind=rp), optional, intent(out)   :: resvec(:)
   real(kind=cp), optional, intent(out)   :: H(:,:)  
    
! Local arrays:
   real(kind=rp)                          :: x(size(b,1),size(b,2))
   real(kind=rp), allocatable             :: P(:,:,:) 
   real(kind=rp), allocatable             :: R0(:,:) 
   real(kind=rp)                          :: G(size(b,1),size(b,2),s)
   real(kind=rp)                          :: U(size(b,1),size(b,2),s)
   real(kind=rp)                          :: r(size(b,1),size(b,2))            
   real(kind=rp)                          :: v(size(b,1),size(b,2))         
   real(kind=rp)                          :: t(size(b,1),size(b,2))        
   real(kind=rp)                          :: M(s,s), f(s), mu(s)
   real(kind=rp)                          :: alpha(s), beta(s), gamma(s) 
    
   real(kind=rp)                          :: om, kappa           
   real(kind=rp)                          :: nr, nt, tr, rho 
    
   include 'idrs_body.f90'
   RIDRS = x
     
   end function RIDRS

end module idrs_module
