program user_main

   write(*,'(a,/,a,/,a,/,a,/,a,/,a,/,a,/,a,/,a,/)') &
      'This programme illustrates IDR(s) using three different test problems:', &
      '   - a complex test problem with a system matrix that is Toeplitz,', &
      '   - a 1D convection-diffusion problem, ', &
      '   - a 2D convection-diffusion problems that is formulated as a matrix equation.', &
      'The three IDR(s) variants are tested:', & 
      '   - Biortho: the default method, this is the most robust version;', &
      '   - BiCGSTAB: IDR(1) with BiCGSTAB parameter choice;', &
      '   - Minsync: a version that is cheaper in memory and operations.', &
      'For Minsync a random initial guess is used to avoid break down.'

    call toeplitz
    call CD_1D
    call CD_2D

    write(*,'(a)') &
      '===================================================================='

end program user_main

subroutine toeplitz

   use user_module
   use idrs_module

   implicit none 
   integer, parameter            :: maxit=4000 
   integer                       :: flag            
   integer                       :: iter          
   integer                       :: s            
  
   integer, parameter            :: N = 200
   type(matrix)                  :: A
   type(preconditioner)          :: M1
   complex(kind=cp)              :: x(N,1), b(N,1), x0(N,1)
   real(kind=rp)                 :: R(N,1), resvec(maxit+1)
   real(kind=rp)                 :: tol, relres
   character*8                   :: method
   integer                       :: i

! Generate the test problems
   A%cind = (/-1, 0, 2, 3/)
   A%cval = (/(0.,3.6d0), (4.d0,0.), (1.d0,0.), (0.7d0,0.)/)

   b = (0.,1.)

! Solve accuracy test
   tol = 1d-12
    write(*,'(a)') &
      '===================================================================='
   write(*,'(a,/,a,/)') &
      'Accuracy test problem, see TOMS paper (2011), example 2', &
      'The termination criterion is ||r||/||b|| < 1e-12. ' 

   method = 'biortho '
   write(*,'(5x,a,/,5x,a)') &
      '   Variant   |    s    | Iterations |  ||r||/||b||  |  Flag  ',&
      '--------------------------------------------------------------'
   do i = 1,7
      s = 2**(i-1)
      x = idrs( A, b, M1, s, tol, maxit, method, flag, relres, iter )
      write(*,'(8x,a8, a,i3,a,i4,a,e9.3,a,i1)') method, &
         '  |  ', s,'    |   ', iter,'     |   ', relres,'   |    ',flag
   end do
   
   method = 'bicgstab'
   s = 1
   x = idrs( A, b, M1, s, tol, maxit, method, flag, relres, iter, resvec=resvec )
   write(*,'(8x,a8, a,i3,a,i4,a,e9.3,a,i1)') method, &
      '  |  ', s,'    |   ', iter,'     |   ', relres,'   |    ',flag

   method = 'minsync '
   call RANDOM_SEED;
   call RANDOM_NUMBER(R)
   x0 = R
   do i = 1,7
      s = 2**(i-1)
      x = idrs( A, b, M1, s, tol, maxit, method, flag, relres, iter, x0 )
      write(*,'(8x,a8, a,i3,a,i4,a,e9.3,a,i1)') method, &
         '  |  ', s,'    |   ', iter,'     |   ', relres,'   |    ',flag
   end do
   write(*,'(5x,a)') &
      '--------------------------------------------------------------'
   write(*,'(18x,a)') 'Results for the complex Toeplitz problem'
   write(*,*)

!
! UNCOMMEND THE CODE BELOW TO COMPARE THE CONVERGENCE WITH MATLAB'S BICGSTAB
!  write(*,'(a)') 'The norms of the residuals of the first ten Bi-CGSTAB iterations are:'
!  do i = 1, 10
!     write(*,'(18x,a,i1,a,e21.15)') '||r_',i-1,'|| = ', resvec(i)
!  end do
!  write(*,*)

end subroutine toeplitz

subroutine CD_1D

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
! 1D convection diffusion test problem to test finite termination
!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

   use user_module
   use idrs_module

   implicit none 
   integer                       :: flag            
   integer, parameter            :: maxit = 4000
   integer                       :: iter          
   integer                       :: s            
  
   real(kind=rp)                 :: h
   real(kind=rp), parameter      :: eps    = 1.d-2
   real(kind=rp), parameter      :: beta_x = 1.
   real(kind=rp), parameter      :: beta_y = 1.
   integer,       parameter      :: m = 80
   integer,       parameter      :: n = 1

   type(matrix)                  :: A
   type(preconditioner)          :: M1
   real(kind=rp)                 :: resvec(maxit+1)
   real(kind=rp)                 :: x(m,n), b(m,n), x0(m,n)
   real(kind=rp)                 :: tol, relres
   character*8                   :: method
   integer                       :: i

! Generate the test problems
   h  = 1.d0/(m+1)
! Define the values of the diagonal elements of the tridiagonal matrices Dx
   A%Dxl = -eps/(h**2) - beta_x/(2*h)
   A%Dxm = 2d0*eps/(h**2)
   A%Dxu = -eps/(h**2) + beta_x/(2*h)

! Define the preconditioner: 
!            precon = 0 means no preconditioning, 
!            precon = 1 means exact inverse (for testing purposes)

   M1%precon = 0 
   M1%Dxl = -0.5d0 - beta_x*h/(4.d0*eps)
   M1%Dxm = 1.d0
   M1%Dxu = -0.5d0 + beta_x*h/(4.d0*eps)
   M1%D = A%Dxm

   b = 0.d0
   b(1,1) = -A%Dxl
   b(m,1) = -A%Dxu

! Solve 1D convection-diffusion test
   write(*,'(a)') &
      '===================================================================='
   tol = 1.d-8
   write(*,'(a,/,a,/,a,/,a,/)') &
      '1D convection-diffusion problem, N=80 gridpoints.', & 
      'This problem tests the finite termination of IDR(s).', &
      'IDR(s) should terminate in at most N+N/s iterations.', &
      'The termination criterion is ||r||/||b|| < 1e-8. ' 

   write(*,'(5x,a,/,5x,a)') &
      '   Variant   |    s    | Iterations |  ||r||/||b||  |  Flag  ',&
      '--------------------------------------------------------------'
   method = 'biortho '
   do i = 1, 4
      s = 2**(i-1)
      x = idrs( A, b, M1, s, tol, maxit, method, flag, relres, iter )
      write(*,'(8x,a8, a,i3,a,i4,a,e9.3,a,i1)') method, &
         '  |  ', s,'    |   ', iter,'     |   ', relres,'   |    ',flag
   end do

   method = 'bicgstab'
   s = 1
   x = idrs( A, b, M1, s, tol, maxit, method, flag, relres, iter, resvec=resvec )
   write(*,'(8x,a8, a,i3,a,i4,a,e9.3,a,i1)') method, &
      '  |  ', s,'    |   ', iter,'     |   ', relres,'   |    ',flag

   method = 'minsync '
   call RANDOM_SEED;
   call RANDOM_NUMBER(x0)
   do i = 1, 4
      s = 2**(i-1)
      x = idrs( A, b, M1, s, tol, maxit, method, flag, relres, iter, x0 )
      write(*,'(8x,a8, a,i3,a,i4,a,e9.3,a,i1)') method, &
         '  |  ', s,'    |   ', iter,'     |   ', relres,'   |    ',flag
   end do
   write(*,'(5x,a)') &
      '--------------------------------------------------------------'
   write(*,'(15x,a)') 'Results for the 1D convection-diffusion problem'
   write(*,*)
!
! UNCOMMEND THE CODE BELOW TO COMPARE THE CONVERGENCE WITH MATLAB'S BICGSTAB
!  write(*,'(a)') 'The norms of the residuals of the first ten Bi-CGSTAB iterations are:'
!  do i = 1, 10
!     write(*,'(18x,a,i1,a,e21.15)') '||r_',i-1,'|| = ', resvec(i)
!  end do
!  write(*,*)

end subroutine CD_1D

subroutine CD_2D

   use user_module
   use idrs_module

   implicit none
   integer                       :: flag
   integer, parameter            :: maxit = 4000
   integer                       :: iter
   integer                       :: s

   real(kind=rp), parameter      :: eps    = 1.e-2
   real(kind=rp), parameter      :: beta_x = 1.
   real(kind=rp), parameter      :: beta_y = 1.
   integer,       parameter      :: m = 80
   integer,       parameter      :: n = 80
   type(matrix)                  :: A
   type(preconditioner)          :: M1
   real(kind=rp)                 :: x(m,n), b(m,n), x0(m,n)
   real(kind=rp)                 :: hx, hy, tol, relres
   integer                       :: i, j
   character*8                   :: method

   hx  = 1./(m+1)
   hy  = 1./(n+1)
! Define the values of the diagonal elements of the tridiagonal matrices
! Dx and Dy
   A%Dxl =  -eps/(hx**2) - beta_x/(2*hx)
   A%Dxm = 2*eps/(hx**2)
   A%Dxu =  -eps/(hx**2) + beta_x/(2*hx)
   A%Dyl =  -eps/(hy**2) - beta_y/(2*hy)
   A%Dym = 2*eps/(hy**2)
   A%Dyu =  -eps/(hy**2) + beta_y/(2*hy)

! Define the preconditioner (simple operator splitting or ADI) 
   M1%precon = 1
   M1%D = A%Dxm+A%Dym
   M1%Dxl = A%Dxl/M1%D
   M1%Dxm = 1.d0
   M1%Dxu = A%Dxu/M1%D
   M1%Dyl = A%Dyl/M1%D
   M1%Dym = 1.d0
   M1%Dyu = A%Dyu/M1%D

! Define the rhs-vector
   b = 0.
   b(:,1) = b(:,1) -A%Dxl
   b(:,n) = b(:,n) -A%Dxu
   b(1,:) = b(1,:) -A%Dyl
   b(m,:) = b(m,:) -A%Dyu

   tol = 1.d-8

! Solve 2D convection-diffusion test
   write(*,'(a)') &
      '===================================================================='
   tol = 1.d-8
   write(*,'(a,/,a,/,a,/,a,/,a,/)') &
      '2D convection-diffusion problem, 80X80 gridpoints.', & 
      'The problem is solved as the matrix equation AX + XB^T = C', &
      'in which A and B are the difference operators in x and y direction. ', &
      'A simple operator splitting (or ADI) preconditioner is used.', &
      'The termination criterion is ||r||/||b|| < 1e-8. ' 

   write(*,'(5x,a,/,5x,a)') &
      '   Variant   |    s    | Iterations |  ||r||/||b||  |  Flag ',&
      '--------------------------------------------------------------'
   method = 'biortho '
   do i = 1, 4
      s = 2**(i-1)
      x = idrs( A, b, M1, s, tol, maxit, method, flag, relres, iter )
      write(*,'(8x,a8, a,i3,a,i4,a,e9.3,a,i1)') method, &
         '  |  ', s,'    |   ', iter,'     |   ', relres,'   |    ',flag
   end do

   method = 'bicgstab'
   s = 1
   x = idrs( A, b, M1, s, tol, maxit, method, flag, relres, iter )
   write(*,'(8x,a8, a,i3,a,i4,a,e9.3,a,i1)') method, &
      '  |  ', s,'    |   ', iter,'     |   ', relres,'   |    ',flag

   method = 'minsync '
   call RANDOM_SEED;
   call RANDOM_NUMBER(x0)
   do i = 1, 4
      s = 2**(i-1)
      x = idrs( A, b, M1, s, tol, maxit, method, flag, relres, iter, x0 )
      write(*,'(8x,a8, a,i3,a,i4,a,e9.3,a,i1)') method, &
         '  |  ', s,'    |   ', iter,'     |   ', relres,'   |    ',flag
   end do
   write(*,'(5x,a)') &
      '--------------------------------------------------------------'
   write(*,'(15x,a)') 'Results for the 2D convection-diffusion problem'
   write(*,*)


end subroutine CD_2D
