module user_module

   implicit none

! Define the precision for real and complex arithmatic

   integer, parameter :: rp = kind(1.d0)                     ! kind for real
   integer, parameter :: cp = kind(1.d0)                     ! kind for complex

! Define the matrix type, as an example here for Toeplitz matrices

   type matrix
      integer                        :: cind(4)          ! Diagonal indices
      complex(kind=cp)               :: cval(4)          ! Diagonal values
      real(kind=rp)                  :: Dxl, Dxm, Dxu, Dyl, Dym, Dyu
   end type matrix

! Overload * to define the matrix-vector multiplication using the matrix type

   INTERFACE OPERATOR(*)
      module procedure rmatvec, cmatvec
   END INTERFACE

! Define the preconditioner type 

   type preconditioner
      integer                        :: precon
      real(kind=rp)                  :: Dxl, Dxm, Dxu, Dyl, Dym, Dyu, D
   end type preconditioner

! Overload / to define the preconditioner operation using the preconditioner type

   INTERFACE OPERATOR(/)
      module procedure rprecon, cprecon
   END INTERFACE


   contains

   function rmatvec( A, v ) 

      type(matrix), intent(in)        :: A
      real(kind=rp), intent(in)       :: v(:,:)
      real(kind=rp)                   :: w(size(v,1),size(v,2))
      real(kind=rp)                   :: rmatvec(size(v,1),size(v,2))
      integer                         :: i,k, m, n
      real(kind=rp)                   :: Dxl, Dxm, Dxu, Dyl, Dym, Dyu, D
      
      m    = size(v,1)
      n    = size(v,2)
      
! Get the values of the diagonal elements of the tridiagonal matrices Dx 
      Dxl = A%Dxl
      Dxm = A%Dxm
      Dxu = A%Dxu
      
      w = 0.
! Compute Dx*v and add to solution
      do k = 1,n
         w(1,k)    = w(1,k) + Dxm*v(1,k)   + Dxu*v(2,k)
         do i = 2,m-1
            w(i,k) = w(i,k) + Dxl*v(i-1,k) + Dxm*v(i,k) + Dxu*v(i+1,k)
         end do
         w(m,k)    = w(m,k) + Dxl*v(m-1,k) + Dxm*v(m,k)
      end do

      if ( n > 1 ) then 

! Get the values of the diagonal elements of the tridiagonal matrices Dy
         Dyl = A%Dyl
         Dym = A%Dym
         Dyu = A%Dyu

! Compute v*Dy^T and add to solution
         do k = 1,m
            w(k,1)    = w(k,1) + Dym*v(k,1)   + Dyu*v(k,2)
            do i = 2,n-1
               w(k,i) = w(k,i) + Dyl*v(k,i-1) + Dym*v(k,i) + Dyu*v(k,i+1)
            end do
            w(k,n)    = w(k,n) + Dyl*v(k,n-1) + Dym*v(k,n)
         end do
      end if
      rmatvec = w

   end function rmatvec

   function cmatvec( A, v ) 

      type(matrix), intent(in)        :: A
      complex(kind=cp), intent(in)    :: v(:,:)
      complex(kind=cp)                :: w(size(v,1),size(v,2))
      complex(kind=cp)                :: cmatvec(size(v,1),size(v,2))
      integer                         :: i,j,k, N, nrhs, ndiag, jdiag
      complex(kind=cp)                :: val

      N    = size(v,1)
      nrhs = size(v,2)
      ndiag = size(A%cind)
      w = 0.
      do k = 1,nrhs
         do j = 1, ndiag
            jdiag = A%cind(j)
            val   = A%cval(j)
            if ( jdiag < 0 ) then
               do i = 1,N+jdiag
                  w(i-jdiag,k) = w(i-jdiag,k) + val*v(i,k)
               end do
            else
               do i = 1,N-jdiag
                  w(i,k) = w(i,k) + val*v(i+jdiag,k)
               end do
            end if
         end do
      end do
      cmatvec = w

   end function cmatvec

   function rprecon( v, M1 )
      implicit none

      type(preconditioner), intent(in)              :: M1
      real(kind=rp), dimension(:,:), intent(in)     :: v
      real(kind=rp), dimension(size(v,1),size(v,2)) :: rprecon
      real(kind=rp), dimension(size(v,1),size(v,2)) :: w
      real(kind=rp)                                 :: Dxl,Dxm,Dxu, D
      real(kind=rp)                                 :: Dyl,Dym,Dyu
      real(kind=rp),dimension(size(v,1))            :: Dxt
      real(kind=rp),dimension(size(v,2))            :: Dyt
      real(kind=rp)                                 :: tmp
      integer                                       :: i, n, m

! Nontrivial preconditioner?
      if ( M1%precon == 0 ) then
         rprecon = v
      else
! Initializations
         m = size(v,1)
         n = size(v,2)
         D   = M1%D
         Dxl = M1%Dxl
         Dxm = M1%Dxm
         Dxu = M1%Dxu

! Forward elimination:
         Dxt(1) = Dxm
         w(1,:) = v(1,:)
         do i = 2,m
            tmp = Dxl/Dxt(i-1)
            Dxt(i) = Dxm   -tmp*Dxu
            w(i,:) = v(i,:)-tmp*w(i-1,:)
         end do

! Backsubstitution:
         w(m,:) = w(m,:)/Dxt(m)
         do i = m-1, 1, -1
            w(i,:) = (w(i,:)-Dxu*w(i+1,:))/Dxt(i)
         end do

         w = w/D

         if ( n > 1 ) then 
            Dyl = M1%Dyl
            Dym = M1%Dym
            Dyu = M1%Dyu

! Forward elimination:
            Dyt(1) = Dym
            do i = 2,n
               tmp = Dyl/Dyt(i-1)
               Dyt(i) = Dym   -tmp*Dyu
               w(:,i) = w(:,i)-tmp*w(:,i-1)
            end do

! Backsubstitution:
            w(:,n) = w(:,n)/Dyt(n)
            do i = n-1, 1, -1
               w(:,i) = (w(:,i)-Dyu*w(:,i+1))/Dyt(i)
            end do
         end if

         rprecon = w
      end if

   end function rprecon

   function cprecon( v, M )

      type(preconditioner), intent(in)      :: M
      complex(kind=cp), intent(in)          :: v(:,:)
      complex(kind=cp)                      :: cprecon(size(v,1),size(v,2))

      cprecon = v

   end function cprecon

end module user_module
